import javax.sound.midi.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

/**
 * Moosique's graphical user interface.
 * 
 * @author  Einar Pehrson
 */
 
public class MooGUI extends JFrame {

	private Sequence seq;
	private MooMenu menu;
	private MooToolbar toolbar;
	private MooView view;
	private JLabel statusBar;
	private java.util.Timer timer;
	private boolean updateView = false;
	public static final int statusResetDelay = 3000;
	public static final Font FONT = new Font("Helvetica", Font.PLAIN, 10);
	public static final Color bgColor = new Color(192, 224, 255);
	public static final Image logo = Toolkit.getDefaultToolkit().getImage("images/moose.gif");
	
	/** 
	 * Creates the GUI.
	 * @param seq The sequence that the program is operating on.
	 */
	public MooGUI(Sequence seq, File file) {
		super("Moosique");
		if (file != null) setTitle("Moosique - " + file.getName());

		this.seq = seq;
		advanceStatus();
		
		Container pane = getContentPane();
		pane.setLayout(new BoxLayout(pane, BoxLayout.Y_AXIS));

		// Adds menu bar.
		menu = new MooMenu();
		setJMenuBar(menu);
		advanceStatus();

		// Adds toolbar.
		toolbar = new MooToolbar();
		pane.add(toolbar, BorderLayout.NORTH);
		advanceStatus();

		// Adds main view.
		view = new MooView(seq.getTracks());
		pane.add(view, BorderLayout.CENTER);

		// Adds status bar.
		statusBar = new JLabel("Welcome to Moosique!", JLabel.CENTER);
		statusBar.setFont(FONT);
		pane.add(statusBar, BorderLayout.SOUTH);

		// Brings on the colors!
		setBackground(pane);
		setBackground(menu);
		setBackground(toolbar);
		setBackground(view);

		// Creates timer.
		timer = new java.util.Timer();

		// Sets up global key listener.
		ActionMap am = getRootPane().getActionMap();

		Action playAction = new AbstractAction() {
			public void actionPerformed(ActionEvent ae) {
				if (!Moosique.getSequencer().isRunning()) {
					Moosique.play();
				} else {
					Moosique.stop();
				}
			}};
		am.put("Play", playAction);
		am.put("Change octave up", createOctaveAction(true));
		am.put("Change octave down", createOctaveAction(false));

		InputMap im = getRootPane().getInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
		KeyStroke playKey = KeyStroke.getKeyStroke(KeyEvent.VK_SPACE, 0);
		KeyStroke octaveUpKey = KeyStroke.getKeyStroke(KeyEvent.VK_F9, 0);
		KeyStroke octaveDownKey = KeyStroke.getKeyStroke(KeyEvent.VK_F10, 0);
		im.put(playKey, "Play");
		im.put(octaveUpKey, "Change octave up");
		im.put(octaveDownKey, "Change octave down");
		advanceStatus();

		// Configures window.
		addWindowListener(new MooGUICloser());
		pack();
		setIconImage(logo);
		Dimension bounds = Toolkit.getDefaultToolkit().getScreenSize();
		setSize(bounds.width,bounds.height - 28);
		setLocation(0, 0);
		setBackground(Color.white);
		advanceStatus();
		setVisible(true);
		show();
	}

	/**
	 * Sets the background on Containers
	 * @param c the Container that will have it's background change
	 */
	private void setBackground(Container c) {
		c.setBackground(bgColor);
		Component[] comps = c.getComponents();
		for (int i = 0; i < comps.length; i++) {
			comps[i].setBackground(bgColor);
		}
	}

	/** 
	 * Changes the sequence of the GUI.
	 * @param sequence	the MIDI sequence to visualize
	 */
	public void setSequence(Sequence sequence, File file) {
		seq = sequence;
		if (file != null) setTitle("Moosique - " + file.getName());
		else setTitle("Moosique");
		view.setTracks(seq.getTracks(), true);
		toolbar.resetProgInd();
	}

	/** 
	 * Shows the given message in the status bar.
	 * @param text	the message to show
	 */
	public void setStatus(String text) {
		statusBar.setText(text);
		timer.schedule(new StatusResetTask(), statusResetDelay);
	}

	/**
	 * Calls on the main view to update the track views,
	 * and on the toolbar to update the progress indicator.
	 */
	public synchronized void update(long tickPosition){
		if (updateView) view.update(tickPosition);
		toolbar.updateProgInd(tickPosition);
	}

	public MooView getView() {
		return view;
	}

	/**
	 * Creates an action for a specific octave.
	 * @param increase 	true for increase, false for decrease
	 */
	private Action createOctaveAction(final boolean increase) {
		Action octaveAction = new AbstractAction() {
			public void actionPerformed(ActionEvent ae) {
				MooKeyboard.setRelativeOctave(increase);
			}};
		return octaveAction;
	}

	/** 
	 * Advances the current progress counter by printing a "." to the System output.
	 */
	private void advanceStatus() {
		System.out.print(".");
	}

	/**
	 * Listener for closing the program
	 */
	class MooGUICloser extends WindowAdapter {
		public void windowClosing(WindowEvent e) {
			Moosique.quit();
		}
	}
	
	/**
	 * TimerTask that resets the statusbar
	 */
	class StatusResetTask extends TimerTask {
		public void run() {
			setStatus(" ");
		}
	}
}
