import javax.sound.midi.*;
import java.io.*;
import javax.swing.*;
import java.util.*;

/**
 * Moosique - The MIDI Tracker
 * 
 * Main class that handles initiation, IO and sound.
 * 
 * @author  Einar Pehrson
 */
 
public class Moosique {

	// GUI and MIDI device variables
	private static MooGUI gui;
	private static Sequence seq;
	private static Sequencer sequencer;
	private static Synthesizer synthesizer;
	private static Receiver receiver;
	private static MidiChannel[] channels;
	private static MidiChannel activeChannel;

	// Recording variables
	private static Track recordTrack = null;
	private static MooTrackView recordTrackView = null;
	private static boolean[] quantizers = {false, false, false};
	private static int quantizeResolution;

	// Collections
	private static ArrayList copyBuffer, emptyTracks, timeSignatures, tempoChanges;
	private static TreeSet selection;
	private static Map trackMute = new HashMap();
	private static Map trackSolo = new HashMap();

	// Various...
	private static File file = null;
	private static long editPosition;
	private static Thread player;

	// Preferences
	private static boolean makeGUI = true, initSound = true, edited = false, drawEmptyTracks = false;

	// Constants
	public static final int DEFAULT_RESOLUTION = 96, DEFAULT_TRACKS = 4;
	public static final int WHOLE_NOTE = 16, HALF_NOTE = 8, QUARTER_NOTE = 4, EIGHTH_NOTE = 2, SIXTEENTH_NOTE = 1;

	/** 
	 * Starts the application.
	 * 
	 * Parses command-line arguments, acquires MIDI devices and connects them,
	 * loads a sequence and creates the GUI.
	 */
	public static void main (String[] args) {
		out("\nMoosique version 1.0\n", true);

		// Parses command-line arguments.
		String fileArg = null;
		for (int i = 0; i < args.length; i++) {
			if (args[i].equals("-n")) makeGUI = false;
			else if (fileArg == null) fileArg = args[i];
		}

		// Acquires MIDI devices and connects them.
		out("Initializing MIDI devices.", false);
		try {
			// Configures sequencer.
			sequencer = MidiSystem.getSequencer();
			advanceStatus();
			sequencer.open();
			sequencer.addMetaEventListener(new SongEndListener());

			// Configures synthesizer.
			synthesizer = MidiSystem.getSynthesizer();
			advanceStatus();
			synthesizer.open();

			// Connects devices.
			receiver = synthesizer.getReceiver();
			sequencer.getTransmitter().setReceiver(receiver);

			// Configures channels.
			channels = synthesizer.getChannels();
			setActiveChannel(0);
		} catch (MidiUnavailableException e) {
			out("Failed, quitting.", true);
			System.exit(1);
		}
		out("Done", true);

		// Loads user preferences (work directory, last opened files etc).
		loadPreferences();

		//If a filename is given as the command-line argument, attempts to load a sequence from the file.
		if (fileArg != null) {
			out("Loading MIDI sequence from " + fileArg + "...", false);
			if (!load(new File(fileArg))) {
				out("Failed, creating new sequence", true);
				clearSequence();
			} else {
				out("Done", true);
			}
		} else {
			// Otherwise creates a new empty one.
			clearSequence();
		}

		// Builds GUI, unless n-flag is set.
		if (makeGUI) {
			out("Building GUI.", false);
			try {
				UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
			} catch (Exception e) {}
			gui = new MooGUI(seq, file);
			out("Done", true);
		} else {
			out("Playing...", false);
			play();
			while (sequencer.isRunning()) {}
			out("Done", true);
			quit();
		}
	}








	/*				***
	 **	  ACCESSOR METHODS	 **
	 ***				  */








	/** 
	 * Returns the currently active MidiChannel.
	 * @return the active MidiChannel
	 */
	public static MidiChannel getActiveChannel() {
		return activeChannel;
	}

	/** 
	 * Returns the MidiChannels of the selected synthesizer.
	 * @return the available MidiChannels
	 */
	public static MidiChannel getChannel(int i) {
		return channels[i];
	}

	/** 
	 * Returns the current copy buffer.
	 * @return the current copy buffer
	 */
	public static ArrayList getCopyBuffer() {
		return copyBuffer;
	}

	/** 
	 * Returns the current editing position of the sequencer.
	 * @return the tick position
	 */
	public static long getEditPosition() {
		return editPosition;
	}

	/** 
	 * Returns the GUI.
	 * @return the GUI
	 */
	public static MooGUI getGUI() {
		return gui;
	}

	/** 
	 * Returns the receiver of the current sequencer.
	 * @return the receiver
	 */
	public static Receiver getReceiver() {
		return receiver;
	}

	/** 
	 * Returns the current sequence.
	 * @return the current sequence
	 */
	public static Sequence getSequence() {
		return seq;
	}

	/** 
	 * Returns the current sequencer.
	 * @return the current sequencer
	 */
	public static Sequencer getSequencer() {
		return sequencer;
	}

	/** 
	 * Returns true if the current sequence has been edited.
	 * @return the tick position
	 */
	public static boolean isEdited() {
		return edited;
	}

	/** 
	 * Returns whether the given track should be drawn
	 * @return true if the given track should be drawn
	 */
	public static boolean shouldBeDrawn(Track track) {
		if (drawEmptyTracks) return true;
		else return (!emptyTracks.contains(track));
	}








	/*				***
	 **	  MUTATOR METHODS	 **
	 ***				  */








	/** 
	 * Fast forwards the current sequence the given number of measures.
	 * @param measures	the number of measures to fast forward
	 */
	public static void forward(long ticks) {
		editPosition += ticks;
	}

	/** 
	 * Rewinds the current sequence the given number of measures.
	 * @param measures	the number of measures to rewind
	 */
	public static void rewind(long ticks) {
		editPosition -= ticks;
	}

	/** 
	 * Sets the currently active MidiChannel.
	 * @param channel	the number of the MidiChannel to activate
	 */
	public static void setActiveChannel(int channel) {
		activeChannel = channels[channel];
	}

	/** 
	 * Sets the current copy buffer.
	 * @param the copy buffer
	 */
	public static void setCopyBuffer(ArrayList buffer) {
		copyBuffer = buffer;
	}

	/** 
	 * Sets whether empty tracks should be drawn
	 * @param state		true if empty tracks should be drawn
	 */
	public static void setDrawEmptyTracks(boolean state) {
		drawEmptyTracks = state;
	}

	/** 
	 * Sets the current sequence as edited, which implies prompts when loading a new sequence.
	 */
	public static void setEdited() {
		edited = true;
	}

	/** 
	 * Sets the current editing position of the sequencer.
	 * @param ticks		the tick position
	 */
	public static void setEditPosition(long ticks) {
		editPosition = ticks;
	}

	/** 
	 * Sets the solo setting of the given track.
	 * @param on	true for solo, false for not
	 */
	public static void setTrackSolo(Track track, boolean on){
		trackSolo.put(track, new Boolean(on));	
	}

	/** 
	 * Sets the mute setting of the given track.
	 * @param on	true for mute, false for not
	 */
	public static void setTrackMute(Track track, boolean on){
		trackMute.put(track, new Boolean(on));	
	}

	/** 
	 * Sets the current playback volume.
	 * @param volume	the volume, between 0 and 1
	 */
	public void setVolume(long volume) {
		for (int i = 0; i < channels.length; i++) {
			channels[i].controlChange(7, (int)(volume * 127.0));
		}
	}








	/*				***
	 **	  SELECTION METHODS	 **
	 ***				  */








	/** 
	 * Returns the current selection.
	 * @return the current selection
	 */
	public static TreeSet getSelection() {
		return selection;
	}

	/**
	 * Selects the given note
	 * @param the note to select
	 */
	public static void selectNote(MooNoteElement elem) {
		selection.add(elem);
	}

	/**
	 * Deselects the given note
	 * @param the note to deselect
	 */
	public static void deselectNote(MooNoteElement elem) {
		selection.remove(elem);
	}

	/**
	 * Deselects all notes.
	 */
	public static void deselectAllNotes() {
		Iterator it = selection.iterator();
		while(it.hasNext()) {
			((MooNoteElement)it.next()).deselect();
		}
		selection.clear();
	}

	/**
	 * Determines if the given MooNoteElement is the only one in the track view that is selected.
	 * @return if the given element is the only selected one
	 */
	public static boolean isTheOnlySelected(MooNoteElement elem) {
		Iterator it = selection.iterator();
		while(it.hasNext()) {
			if (!it.next().equals(elem)) return false;
		}
		return true;
	}








	/*				***
	 **	  PLAYBACK METHODS	 **
	 ***				  */








	/** 
	 * Starts playback of the current sequence.
	 */
	public static void play() {
		sequencer.setTickPosition(editPosition);
		resume();
	}

	/** 
	 * Resumes playback of the current sequence.
	 */
	public static void resume() {
		gui.update(0);
		try {
			sequencer.setSequence(seq);
		} catch (InvalidMidiDataException e) {}
		Track[] tracks = seq.getTracks();

		sequencer.start();

		for (int i = 0; i < tracks.length; i++) {

			Object ob = trackSolo.get(tracks[i]);
			if(ob instanceof Boolean){
				sequencer.setTrackSolo(i,((Boolean)ob).booleanValue());
			}

			ob = trackMute.get(tracks[i]);
			if(ob instanceof Boolean){
				sequencer.setTrackMute(i,((Boolean)ob).booleanValue());
			}
		}

		// Disables input to volatile components
		// gui.disable();

		// Creates the visualisation thread and starts it.
		player = new PlayThread();
		player.start();
	}

	/** 
	 * Pauses playback of the current sequence.
	 */
	public static void pause() {
		if (sequencer.isRunning()) {
			sequencer.stop();
		}
		player = null;
	}

	/** 
	 * Stops playback of the current sequence.
	 */
	public static void stop() {
		if (sequencer.isRunning()) {
			sequencer.stop();
		}
		sequencer.setTickPosition(editPosition);
		player = null;
		gui.update((long)0);

		if (recordTrack != null && recordTrackView != null) {
			recordTrackView.disableKeyboardRecording();
			sequencer.stopRecording();
			sequencer.recordDisable(recordTrack);
			if (quantizers[0]) recordTrackView.placeNewNotes(quantize(
				convertTrack(recordTrack), quantizeResolution, quantizers[1], quantizers[2]));
			else recordTrackView.placeNewNotes(Moosique.convertTrack(recordTrack));
		}
	}

	/** 
	 * Enables recording to the given track.
	 * @param track		the track in which to store the recorded data
	 * @param tempo		the channel from which to record data
	 * @param quantizers	an array of booleans where 0 = quantize?, 1 = location, 2 = duration
	 * @param resolution	the note size to round each note to
	 * @return if the recording was initialised successfully
	 */
	public static boolean record(Track track, int channel, boolean[] quants, int resolution) {
		try {
			sequencer.recordEnable(track, channel);
			sequencer.startRecording();
		} catch(Exception e) {
			e.printStackTrace();
			return false;
		}
		quantizers = quants;
		quantizeResolution = resolution;
		recordTrack = track;
		recordTrackView = gui.getView().getTrackView(track);
		recordTrackView.enableKeyboardRecording();
		Moosique.setEdited(); 
		return true;
	}








	/*				    ***
	 **  TEMPO & TIME SIGNATURE METHODS  **
	 ***				      */








	/** 
	 * Returns the byte array for the given tempo.
	 * @param tempo	the tempo in beats per minute
	 * @return an array of bytes representing the given tempo
	 */
	public static byte[] encodeTempo(int tempo) {
		int microSecsPerQuarter = 60000000 / tempo;
		byte[] b = new byte[3];
		b[0] = (byte)(microSecsPerQuarter / 65536);
		b[1] = (byte)((microSecsPerQuarter - (b[0] * 65536)) / 256);
		b[2] = (byte)(microSecsPerQuarter - (b[0] * 65536) - (b[1] * 256));
		return b;
	}

	/** 
	 * Returns the tempo for the given byte array.
	 * @param an array of three bytes representing the tempo
	 * @return the tempo in beats per minute
	 */
	public static int decodeTempo(byte[] bytes) {
		return 60000000 / (bytes[0] * 65536 + bytes[1] * 256 + bytes[2]);
	}

	/** 
	 * Returns the tempo in the given tick position.
	 * @param tick the tick position for which to return the tempo
	 * @return the tempo at the specified tick position
	 */
	public static int getTempo(long tick) {
		if (tempoChanges.size() == 0) return 120;
		MidiEvent tempoEvent = (MidiEvent)tempoChanges.get(0);
		if (tempoChanges.size() > 1) {
			for (int i = 1; i < tempoChanges.size(); i++) {
				MidiEvent nextTempoEvent = (MidiEvent)tempoChanges.get(i);
				if (nextTempoEvent.getTick() < tick && nextTempoEvent.getTick() > tempoEvent.getTick())
					tempoEvent = nextTempoEvent;
			}
		}
		return decodeTempo(((MetaMessage)tempoEvent.getMessage()).getData());
	}

	/** 
	 * Sets the tempo at the given tick position.
	 * @param ticks		the tick position
	 */
	public static void setTempo(long tick, int bpm) {
		// Checks for a tempo event at the given tick position. 
		MidiEvent tempoEvent = null;
		Iterator it = tempoChanges.iterator();
		while(it.hasNext()) {
			MidiEvent nextTempoEvent = (MidiEvent)it.next();
			if (nextTempoEvent.getTick() == tick) {
				tempoEvent = nextTempoEvent;
				break;
			}
		}

		// If none was found, creates and adds a new one.
		if (tempoEvent == null) {
			tempoEvent = new MidiEvent(new MetaMessage(), tick);
			(seq.getTracks())[0].add(tempoEvent);
			tempoChanges.add(tempoEvent);
			Collections.sort(tempoChanges, new MidiEventComparator());
		}

		// Sets the tempo of the event (found or created).
		try {
			((MetaMessage)tempoEvent.getMessage()).setMessage(81, encodeTempo(bpm), 3);
		} catch (InvalidMidiDataException e) {}
	}

	/** 
	 * Returns the byte array for the given time signature.
	 * @param numerator	the numerator of the time signature
	 * @param denominator	the denominator of the time signature
	 * @return an array of bytes representing the given time signature
	 */
	public static byte[] encodeTimeSig(int numerator, int denominator) {
		byte[] b = {
			(byte)numerator,
			(byte)(Math.log(denominator) / Math.log(2)), // logarithm of denominator in base 2
			(byte)96,
			(byte)8
		};
		return b;
	}
	/** 
	 * Returns the time signature for the given byte array.
	 * @param an array of four bytes representing the time signature
	 * @return an array of two integers where [0] is the numerator and [1] the denominator
	 */
	public static int[] decodeTimeSig(byte[] bytes) {
		int[] t = {
			(int)bytes[0],
			(int)(1 << bytes[1])
		};
		return t;
	}

	/** 
	 * Returns the time signature in the given tick position.
	 * @param tick the tick position for which to return the time signature
	 * @return an array of two integers where [0] is the numerator and [1] the denominator
	 */
	public static int[] getTimeSig(long tick) {
		int[] ts = {4, 4};
		if (timeSignatures.size() == 0) return ts;
		MidiEvent timeSigEvent = (MidiEvent)timeSignatures.get(0);
		if (timeSignatures.size() > 1) {
			for (int i = 1; i < timeSignatures.size(); i++) {
				MidiEvent nextTimeSigEvent = (MidiEvent)timeSignatures.get(i);
				if (nextTimeSigEvent.getTick() <= tick && nextTimeSigEvent.getTick() > timeSigEvent.getTick())
					timeSigEvent = nextTimeSigEvent;
			}
		}
		return decodeTimeSig(((MetaMessage)timeSigEvent.getMessage()).getData());
	}

	/** 
	 * Sets the time signature at the given tick position.
	 * @param ticks		the tick position
	 */
	public static void setTimeSig(long tick, int numerator, int denominator) {
		// Checks for a time signature event at the given tick position. 
		MidiEvent timeSigEvent = null;
		Iterator it = timeSignatures.iterator();
		while(it.hasNext()) {
			MidiEvent nextTimeSigEvent = (MidiEvent)it.next();
			if (nextTimeSigEvent.getTick() == tick) {
				timeSigEvent = nextTimeSigEvent;
				break;
			}
		}

		// If none was found, creates and adds a new one.
		if (timeSigEvent == null) {
			timeSigEvent = new MidiEvent(new MetaMessage(), tick);
			(seq.getTracks())[0].add(timeSigEvent);
			timeSignatures.add(timeSigEvent);
			Collections.sort(timeSignatures, new MidiEventComparator());
		}

		// Sets the time signature of the event (found or created).
		try {
			((MetaMessage)timeSigEvent.getMessage()).setMessage(88, encodeTimeSig(numerator, denominator), 4);
		} catch (InvalidMidiDataException e) {}
	}
	
	/**
	 * Calculates the position (measures, beats, ticks) in the current sequence for the given tick position.
	 * @param tickPosition the tick position for which to calculate the position
	 * @return an array of integers where index 0 is measures, 1 is beats and 2 is ticks.
	 */
	public static int[] getPositionForTicks(long tickPosition) {
		int ticksPerBeat = seq.getResolution();
		long measures = 0, beats = 0, ticks = 0;

		// Counts for each time signature change up to the last one before the given tick position.
		Iterator it = timeSignatures.iterator();
		MidiEvent lastTSEvent = (MidiEvent)it.next();
		while(it.hasNext()) {
			MidiEvent nextTSEvent = (MidiEvent)it.next();
			if (nextTSEvent.getTick() > tickPosition) break;
			long tickDiff = nextTSEvent.getTick() - lastTSEvent.getTick();
			int[] ts = decodeTimeSig(((MetaMessage)lastTSEvent.getMessage()).getData());
			int beatsPerMeasure = ts[0] * (4 / ts[1]);
			long thisTSMeasures = tickDiff / (beatsPerMeasure * ticksPerBeat);
			measures += thisTSMeasures;
			long thisTSBeats = (tickDiff - thisTSMeasures * beatsPerMeasure * ticksPerBeat) / ticksPerBeat;
			beats += thisTSBeats;
			ticks += tickDiff - thisTSMeasures * beatsPerMeasure * ticksPerBeat - thisTSBeats * ticksPerBeat;
			lastTSEvent = nextTSEvent;
		}

		// Counts from the given tick position to the last time signature change before it.
		long tickDiff = tickPosition - lastTSEvent.getTick();
		int[] ts = decodeTimeSig(((MetaMessage)lastTSEvent.getMessage()).getData());
		int beatsPerMeasure = ts[0] * (4 / ts[1]);
		long thisTSMeasures = tickDiff / (beatsPerMeasure * ticksPerBeat);
		measures += thisTSMeasures;
		long thisTSBeats = (tickDiff - thisTSMeasures * beatsPerMeasure * ticksPerBeat) / ticksPerBeat;
		beats += thisTSBeats;
		ticks += tickDiff - thisTSMeasures * beatsPerMeasure * ticksPerBeat - thisTSBeats * ticksPerBeat;

		// Corrects any overflows.
		if (ticks > ticksPerBeat) {
			beats += Math.floor(ticks / ticksPerBeat);
			ticks = ticks % ticksPerBeat;
		}
		if (beats > beatsPerMeasure) {
			measures += Math.floor(beats / beatsPerMeasure);
			beats = beats % beatsPerMeasure;
		}

		// Returns the calculated values.
		int[] pos = {(int)measures, (int)beats, (int)ticks};
		return pos;
	}

	/** 
	 * Calculates the tick position in the current sequence for the given position (measures, beats, ticks).
	 * @param measures	the measure of the current position
	 * @param beats		the beat of the current position
	 * @param tick		the tick of the current position
	 * @return the tick position.
	 */
	public static long getTicksForPosition(int measures, int beats, int ticks) {
		int res = seq.getResolution();
		int[] lastTSPos = new int[3];
		long tickPosition = 0;

		// Counts for each time signature change up to the last one before the given tick position.
		Iterator it = timeSignatures.iterator();
		MidiEvent lastTSEvent = (MidiEvent)it.next();
		while(it.hasNext()) {
			MidiEvent nextTSEvent = (MidiEvent)it.next();
			int[] nextTSPos = getPositionForTicks(nextTSEvent.getTick());
			if (nextTSPos[0] >= measures) break;
			lastTSPos = nextTSPos;
			lastTSEvent = nextTSEvent;
		}

		// Counts from the given tick position to the last time signature change before it.
		int measdiff = measures - lastTSPos[0];
		int[] ts = decodeTimeSig(((MetaMessage)lastTSEvent.getMessage()).getData());
		int beatsPerMeasure = ts[0] * (4 / ts[1]);
		tickPosition = lastTSEvent.getTick() + (beatsPerMeasure * measures + beats) * res + ticks;

		return tickPosition;
	}








	/*				***
	 **	  SYSTEM & IO METHODS	 **
	 ***				  */








	/** 
	 * Replaces the current sequence with a new one, holding three empty tracks.
	 */
	public static void clearSequence() {
		// Reinitializes sequence variables
		file = null;
		reinitializeLists();

		try {
			// Creates a new sequence.
			seq = new Sequence(Sequence.PPQ, DEFAULT_RESOLUTION, DEFAULT_TRACKS);
			Track[] tracks = seq.getTracks();

			// Sets default tempo (120) and time signature (4/4) at the beginning of the sequence.
			setTempo(0, 120);
			setTimeSig(0, 4, 4);

			// Sets program and title for the tracks.
			initializeTrack(tracks[1], 0, 24, "Guitar");
			initializeTrack(tracks[2], 1, 33, "Bass");
			initializeTrack(tracks[3], 9, 0, "Drums");
		} catch (InvalidMidiDataException e) {}

		// Sends the sequence to the GUI.
		if (gui != null) gui.setSequence(seq, null);
	}

	/** 
	 * Creates event in the given track for program change and title.
	 */
	private static void initializeTrack(Track track, int channel, int program, String title) {
			// Creates program change and title message.
			ShortMessage programMsg = new ShortMessage();
			MetaMessage titleMsg = new MetaMessage();

			// Sets the data of the messages.
			try {
				programMsg.setMessage(ShortMessage.PROGRAM_CHANGE, channel, program, 0);
				titleMsg.setMessage(3, title.getBytes(), title.length());
			} catch (InvalidMidiDataException e) {}

			// Sends the program change to the channel
			getChannel(channel).programChange(program);

			// Adds them to the track.
			track.add(new MidiEvent(programMsg, (long)0));
			track.add(new MidiEvent(titleMsg, (long)0));
	}
	
	/** 
	 * Wraps each NoteOn event in the track with its NoteOff event in a MooNote.
	 * @param track		the track to convert
	 * @param quantize	whether to round locations and durations in the track to nearest 16th
	 * @return a list of the created MooNotes
	 */
	public static List convertTrack(Track track) {
		// Searches the track for NoteOn and NoteOff events
		ArrayList noteOns = new ArrayList(track.size() / 2);
		ArrayList noteOffs = new ArrayList(track.size() / 2);
		ArrayList newMooNotes = new ArrayList();
		MidiEvent event;
		for (int j = 0; j < track.size(); j++) {
			event = track.get(j);
			if (event.getMessage().getStatus() >= 144 &&
			    event.getMessage().getStatus() < 160) noteOns.add(event);
			if (event.getMessage().getStatus() >= 128 &&
			    event.getMessage().getStatus() < 144) noteOffs.add(event);
		}
		noteOns.trimToSize();
		noteOffs.trimToSize();
		if (noteOns.size() == 0) emptyTracks.add(track);
		
		// Sorts the note lists by tick position.
		Comparator c = new MidiEventComparator();
		Collections.sort(noteOns, c);
		Collections.sort(noteOffs, c);

		// Replaces each NoteOn event it with a MooNote containing a reference to the NoteOff event.
		Iterator iOn = noteOns.iterator(), iOff;
		MidiEvent on, off = null, nextOff;
		ShortMessage onMsg, nextOffMsg;
		while(iOn.hasNext()) {
			on = (MidiEvent)iOn.next();
		    	if (!(on instanceof MooNote)) {
				onMsg = (ShortMessage)on.getMessage();
				iOff = noteOffs.iterator();
				while(iOff.hasNext()) {
					nextOff = (MidiEvent)iOff.next();
					nextOffMsg = (ShortMessage)nextOff.getMessage();
					if(onMsg.getChannel() == nextOffMsg.getChannel() &&
					   onMsg.getData1() == nextOffMsg.getData1() &&
					   c.compare(nextOff, on) > 0) {
						off = nextOff;
						iOff.remove();
						break;
					}
					   	
				}
				track.remove(on);
				MooNote mn;
				if (off != null) {
					mn = new MooNote(on, off);
				} else {
					mn = new MooNote(on, new MidiEvent((ShortMessage)on.getMessage().clone(), on.getTick() + 48));
				}
				track.add(mn);
				newMooNotes.add(mn);
				iOn.remove();
			}
		}
		return newMooNotes;
	}

	/** 
	 * Loads a MIDI sequence from the given file.
	 * @param filename	the filename to use
	 */
	public static boolean load(File loadFile) {
		// Loads sequence from file
		try {seq = MidiSystem.getSequence(loadFile);}
		catch (Exception e) {return false;}
		file = loadFile;
		edited = false;

		Track[] tracks = seq.getTracks();
		reinitializeLists();

		// Searches track 0 for changes in tempo and time signature.
		MidiEvent event;
		MetaMessage metaMsg;
		for (int i = 0; i < tracks[0].size(); i++) {
			event = tracks[0].get(i);
			if (event.getMessage().getStatus() == MetaMessage.META) {
				metaMsg = (MetaMessage)event.getMessage();
				switch(metaMsg.getType()) {
					case 81: tempoChanges.add(event); break;
					case 88: timeSignatures.add(event);
				}
			}
		}
		Comparator c = new MidiEventComparator();
		Collections.sort(tempoChanges, c);
		Collections.sort(timeSignatures, c);

		try {
			// If no time signature specified at tick 0, adds the standard one.
			if (timeSignatures.size() == 0 || ((MidiEvent)timeSignatures.get(0)).getTick() != 0) {
				MetaMessage timeSigMsg = new MetaMessage();
				timeSigMsg.setMessage(88, encodeTimeSig(4, 4), 4);
				timeSignatures.add(0, new MidiEvent(timeSigMsg, (long)0));
			}

			// If no tempo specified at tick 0, adds the standard one.
			if (tempoChanges.size() == 0 || ((MidiEvent)tempoChanges.get(0)).getTick() != 0) {
				MetaMessage tempoMsg = new MetaMessage();
				tempoMsg.setMessage(81, encodeTempo(120), 3);
				tempoChanges.add(0, new MidiEvent(tempoMsg, (long)0));
			}
		} catch (Exception e) {}

		// Converts tracks.
		for (int i = 0; i < tracks.length; i++) {
			convertTrack(tracks[i]);
		}

		// Sends sequence to GUI and sequencer, then returns
		if (gui != null) gui.setSequence(seq, file);
		try {
			sequencer.setSequence(seq);
		} catch (InvalidMidiDataException e) {}
		return true;
	}

	/** 
	 * Quantizes the given list of MIDI events
	 * @param notes		a list of the notes to quantize
	 * @param resolution	the note size to round each note to
	 * @param location	whether the quantize should affect the location of the note
	 * @param duration	whether the quantize should affect the duration of the note
	 */
	public static List quantize(List notes, int resolution, boolean location, boolean duration) {
		Iterator it = notes.iterator();
		int noteSize = resolution * seq.getResolution() / 4;
		while(it.hasNext()) {
			MidiEvent note = (MidiEvent)it.next();
			if (note instanceof MooNote) {
				MooNote mn = (MooNote)note;
				if (location) mn.setTick(Math.round(mn.getTick() / noteSize) * noteSize);
				if (duration) {
					int length = Math.round(mn.getDuration() / noteSize) * noteSize;
					if (length < noteSize) length = noteSize;
					mn.setDuration(length);
				}
			}
		}
		return notes;
	}

	/** 
	 * Reinitializes sequence-specific collections.
	 */
	private static void reinitializeLists() {
		emptyTracks	= new ArrayList();
		timeSignatures 	= new ArrayList();
		tempoChanges 	= new ArrayList();
		copyBuffer 	= new ArrayList();
		trackSolo 	= new HashMap();
		trackMute 	= new HashMap();
		selection 	= new TreeSet();
	}

	/** 
	 * Loads the user preferences.
	 */
	public static void loadPreferences() {
		
	}

	/** 
	 * Saves the user preferences.
	 */
	public static void savePreferences() {
		
	}

	/** 
	 * Prompts the user to save any unsaved changes.
	 */
	public static boolean promptOnUnsavedChanges() {
		if (!edited) return false;
		int exitOption = JOptionPane.showConfirmDialog(gui,
			"The current sequence has been edited, but not saved.\nDo you wish to continue anyway?",
			"File not saved - continue?", 
			JOptionPane.OK_CANCEL_OPTION, 
			JOptionPane.WARNING_MESSAGE);
		if (exitOption == JOptionPane.CANCEL_OPTION || exitOption == JOptionPane.CLOSED_OPTION) return true;
		return false;
	}

	/** 
	 * Saves the current sequence to the previously given filename.
	 */
	public static boolean save() {
		if (file == null) return false;
		else {
			saveAs(file);
			return true;
		}
	}

	/** 
	 * Saves the current sequence to the given filename
	 * @param file	the filename to use
	 */
	public static boolean saveAs(File saveFile) {
		try {
			MidiSystem.write(seq, 1, saveFile);
			file = saveFile;
			edited = false;
			gui.setStatus("Saved " + file.getAbsolutePath());
			return true;
		} catch (IOException e) {
			gui.setStatus("Failed in saving " + file.getAbsolutePath());
			return false;
		}
	}

	/** 
	 * Prints the given string to the System output.
	 */
	private static void out(String text, boolean newLine) {
		if (newLine) System.out.println(text);
		else System.out.print(text);
	}

	/** 
	 * Advances the current progress counter by printing a "." to the System output.
	 */
	private static void advanceStatus() {
		out(".", false);
	}

	/** 
	 * Releases all reserved devices and exits the program.
	 */
	public static void quit() {
		if (gui != null) {
			if (promptOnUnsavedChanges()) return;
		}
		savePreferences();
		if (sequencer.isOpen()) sequencer.close();
		if (synthesizer.isOpen()) synthesizer.close();
		System.exit(0);
	}
	
	/** 
	 * A Ccmparator for sorting lists of MidiEvents.
	 */
	public static class MidiEventComparator implements Comparator {
		public int compare(Object o1, Object o2) {
			int diff = (int)(((MidiEvent)o1).getTick() - ((MidiEvent)o2).getTick());
			if (diff != 0 || !(((MidiEvent)o1).getMessage() instanceof ShortMessage) || !(((MidiEvent)o2).getMessage() instanceof ShortMessage)) return diff;
			return (((ShortMessage)((MidiEvent)o1).getMessage()).getData1() - ((ShortMessage)((MidiEvent)o2).getMessage()).getData1());
		}
	}

	/** 
	 * The thread that updates the GUI during playback.
	 */
	public static class PlayThread extends Thread {
		public void run() {
			Thread currentThread = Thread.currentThread();
			while(currentThread == player) {
				gui.update(sequencer.getTickPosition());
				try {
					sleep(10);
				} catch (InterruptedException e) {
					Moosique.stop();
				}
			}
		}
	}

	/** 
	 * A listener for detecting the Meta event signifying the end of the sequence.
	 */
	public static class SongEndListener implements MetaEventListener {
		public void meta(MetaMessage event) {
			if (event.getType() == 47)
				stop();
 		}
	}
}
